# based on code written by GPT-4
import csv
import os

from pywikibot import Claim, WbTime, ItemPage, PropertyPage, Site
from datetime import datetime
import mwclient
from dotenv import load_dotenv

load_dotenv()


def claim_to_string(claim):
    subject_qid = claim.on_item.id
    predicate_pid = claim.getID()

    # Object QID, assuming the target is a Wikidata item
    # Note: This simplification assumes the claim's target is an item.
    # For other target types (e.g., quantities, strings), additional handling is needed.
    if isinstance(claim.getTarget(), ItemPage):
        object_qid = claim.getTarget().id
    else:
        # Placeholder or additional logic for non-ItemPage targets
        object_qid = 'N/A'  # This could be expanded to handle other types of targets

    return f"({subject_qid})-[{predicate_pid}]-({object_qid})"


# Function to check if a specific time qualifier exists
def time_qualifier_exists(claim, qualifier_pid, year_value):
    for qualifier in claim.qualifiers.get(qualifier_pid, []):
        qualifier_date = qualifier.getTarget()
        if qualifier_date.year == year_value:
            print(f'Time qualifier {qualifier_pid} with value {year_value} already exists on {claim_to_string(claim)}.')
            return True
    return False


def add_time_qualifiers(repo, claim, start_time, end_time):
    qualifiers = []

    if (start_time and end_time) and (start_time == end_time):
        if not time_qualifier_exists(claim, 'P585', int(start_time)):
            point_in_time_qualifier = Claim(repo, 'P585')
            point_in_time_qualifier.setTarget(WbTime(year=int(start_time)))
            claim.addQualifier(point_in_time_qualifier, summary='Adding point in time')
            print(f'Added point_in_time qualifier to {claim_to_string(claim)}')
            qualifiers.append(point_in_time_qualifier)

    else:
        if start_time and not time_qualifier_exists(claim, 'P580', int(start_time)):
            start_time_qualifier = Claim(repo, 'P580')
            start_time_qualifier.setTarget(WbTime(year=int(start_time)))
            claim.addQualifier(start_time_qualifier, summary='Adding start time')
            print(f'Added start_time qualifier to {claim_to_string(claim)}')
            qualifiers.append(start_time_qualifier)

        if end_time and not time_qualifier_exists(claim, 'P582', int(end_time)):
            end_time_qualifier = Claim(repo, 'P582')
            end_time_qualifier.setTarget(WbTime(year=int(end_time)))
            claim.addQualifier(end_time_qualifier, summary='Adding end time')
            print(f'Added end_time qualifier to {claim_to_string(claim)}')
            qualifiers.append(end_time_qualifier)

    return qualifiers


# Function to check if a reference with the given URL already exists on the claim
def reference_url_exists(claim, url):
    for source in claim.getSources():
        if 'P4656' in source or 'P854' in source:  # Check both Wikimedia import URL and reference URL
            for prop in source.get('P4656', []) + source.get('P854', []):
                if prop.getTarget() == url:
                    print(f'Source URL {url} already exists on {claim_to_string(claim)}.')
                    return True
    return False


def qualifier_exists(claim, qualifier_property_id, target):
    for existing_qualifier in claim.qualifiers.get(qualifier_property_id, []):
        if existing_qualifier.getTarget() == target:
            print(
                f'Qualifier {qualifier_property_id} with value {target.getID()} already exists on {claim_to_string(claim)}.')
            return True
    return False


def add_reference(repo, claim, reference_url, retrieved_at_time, qualifiers=None):
    sources = []
    if reference_url and not reference_url_exists(claim, reference_url):
        # Determine whether the URL is a Wikipedia URL or another type of URL
        property_id = 'P4656' if 'wikipedia.org' in reference_url else 'P854'

        # Create the reference claim
        source_claim = Claim(repo, property_id)
        source_claim.setTarget(reference_url)
        sources.append(source_claim)

        # Create the 'retrieved at' claim
        retrieved_at_claim = Claim(repo, 'P813')
        retrieved_at_target = WbTime(year=retrieved_at_time.year, month=retrieved_at_time.month,
                                     day=retrieved_at_time.day)
        retrieved_at_claim.setTarget(retrieved_at_target)
        sources.append(retrieved_at_claim)

    # If a qualifier has been passed for which this reference is the source, add it
    if qualifiers:
        for qualifier in qualifiers:
            supports_qualifier_claim = Claim(repo, 'P10551')  # "supports qualifier"
            site = Site("wikidata", "wikidata")
            property_page = PropertyPage(site, qualifier.getID())
            if not qualifier_exists(claim, 'P10551', property_page):
                supports_qualifier_claim.setTarget(property_page)
                sources.append(supports_qualifier_claim)

    # Add the references to the claim
    if len(sources) > 0:
        claim.addSources(sources, summary='Adding reference and retrieved at date')
        print(f'Added references to {claim_to_string(claim)}')

    return sources


# main function
def update_wikidata_from_csv(file_path):
    site = Site("wikidata", "wikidata")
    repo = site.data_repository()

    previous_object_qid = None
    previous_claim = None

    with open(file_path, newline='', encoding='utf-8') as csvfile:
        reader = csv.DictReader(csvfile)
        for row in reader:
            print("----------")
            subject_qid = row['subject-qid']
            pid = row['pid']
            object_qid = row['object-qid']
            start_time = row['start_time']
            end_time = row['end_time']
            reference_url = row['reference_url']

            # If the new subject is identical to the old object, refine the previous claim
            if subject_qid == previous_object_qid and previous_claim:
                claim = previous_claim
                print(f'Refining {claim_to_string(claim)}')
            else:
                item = ItemPage(repo, subject_qid)
                item.get()

                # Check if the claim already exists
                claim_exists = False
                for claim in item.claims.get(pid, []):
                    if claim.getTarget().id == object_qid:
                        claim_exists = True
                        print(f'{claim_to_string(claim)} exists.')
                        break

                if not claim_exists:
                    claim = Claim(repo, pid)
                    target = ItemPage(repo, object_qid)
                    claim.setTarget(target)
                    item.addClaim(claim)
                    print(f'Created {claim_to_string(claim)}')

            # start_time and end_time
            qualifiers = add_time_qualifiers(repo, claim, start_time, end_time)

            # references
            retrieved_at_time = datetime.utcnow()
            add_reference(repo, claim, reference_url, retrieved_at_time, qualifiers)

            # Remember the object and claim for the next iteration
            previous_object_qid = object_qid
            previous_claim = claim


def get_wikipedia_page_data(pageTitle: str, language="en"):
    user_agent = os.getenv('USER_AGENT')
    site = mwclient.Site(f'{language}.wikipedia.org', clients_useragent=user_agent)
    page = site.pages[pageTitle]
    if not page.exists:
        return None

    return {
        'page': page,
        'revision': page.revision,
        'url': f'{site.host}/wiki/{pageTitle.replace(" ", "_")}?oldid={page.revision}',
        'content': page.text()
    }